from ..serializers.base_serializer import BaseSerializer
from ..pubsub_providers.model_channel_builder import make_channels
from ..tests.serializers import CompanySerializer
from .dragon_test_case import DragonTestCase
from .mock_connection import TestConnection


class TestObject(object):
    def __init__(self, name, numval):
        super(TestObject, self).__init__()
        self.name = name
        self.numval = numval


class TestObjectSerializer(BaseSerializer):
    publish_fields = ['name', 'numval']
    update_fields = []


class PubSubTest(DragonTestCase):
    def test_subscribe_to_all_test_objects(self):
        connection = TestConnection()
        channel = 'test_chan'
        connection.pub_sub.subscribe(channel, connection)
        test_object = TestObject('foo', 1)
        connection.pub_sub.publish(channel, TestObjectSerializer().serialize(test_object))
        test_object = TestObject('bar', 2)
        connection.pub_sub.publish(channel, TestObjectSerializer().serialize(test_object))
        self.assertEqual(len(connection.pub_sub._subscribers.keys()), 1)

    def test_and_or_channel_filters(self):
        context_a = {
            'id': 1,
            'name': 'test'
        }
        context_b = [{'id': 1, }, {'name': 'test'}]
        channels_a = make_channels(CompanySerializer, property_filter=context_a)
        channels_b = make_channels(CompanySerializer, property_filter=context_b)
        self.assertEqual(channels_a, ['company|id:1|name:test'])
        self.assertListEqual(channels_b, [
            'company|id:1',
            'company|name:test'
        ])



from .dragon_django_test_case import DragonDjangoTestCase
from .serializers import FooSerializer, BarSerializer, BazSerializer, QuxSerializer, CompanySerializer, DepartmentSerializer, StaffSerializer
from .models import Company, CompanyLogo, Department, Staff


class ObjectMapTest(DragonDjangoTestCase):
    def test_get_object_map(self):
        foo_graph = FooSerializer.get_object_map()
        bar_graph = BarSerializer.get_object_map()
        baz_graph = BazSerializer.get_object_map()
        qux_graph = QuxSerializer.get_object_map()

        company_graph = CompanySerializer.get_object_map()
        dep_graph = DepartmentSerializer.get_object_map()
        staff_graph = StaffSerializer.get_object_map()

        company = Company.objects.create(name='test co', comp_num=1)
        dep = Department.objects.create(company=company)



import json
from .. import route_handler
from ..route_handler import BaseRouter
from ..permissions import login_required, LoginRequired
from .mock_connection import TestConnection
from .models import TestUser
from .dragon_django_test_case import DragonDjangoTestCase


class PermissionRouter(BaseRouter):
    route_name = 'test_permission_attr_router'
    valid_verbs = ['test_perm']

    @login_required
    def test_perm(self, **kwargs):
        self.send('foo')


class PermissionClassRouter(BaseRouter):
    route_name = 'test_permission_class_router'
    valid_verbs = ['test_perm']
    permission_classes = [LoginRequired()]

    def test_perm(self, **kwargs):
        self.send('foo')


class TestPermissions(DragonDjangoTestCase):
    def setUp(self):
        route_handler.register(PermissionRouter)
        self.permission_handler = route_handler.get_route_handler(PermissionRouter.route_name)

    def test_permission_decorator_anonymous_user(self):
        data = {'verb': 'test_perm'}
        connection = TestConnection()
        self.permission_handler(connection).handle(data)
        json_data = json.loads(connection.sent_data[-1])
        self.assertEqual(json_data['context']['state'], 'login_required')
        self.assertNotEqual(json_data['data'], 'foo')

    def test_permission_decorator_user(self):
        data = {'verb': 'test_perm'}
        connection = TestConnection(user=TestUser(username='testuser', id=1))
        self.permission_handler(connection).handle(data)
        json_data = json.loads(connection.sent_data[-1])
        self.assertEqual(json_data['context']['state'], 'success')
        self.assertEqual(json_data['data'], 'foo')


class TestClassPermissions(DragonDjangoTestCase):
    def setUp(self):
        route_handler.register(PermissionClassRouter)
        self.permission_handler = route_handler.get_route_handler(PermissionClassRouter.route_name)

    def test_permission_dectorator_anonymous_user(self):
        data = {'verb': 'test_perm'}
        connection = TestConnection()
        self.permission_handler(connection).handle(data)
        json_data = json.loads(connection.sent_data[-1])
        self.assertEqual(json_data['context']['state'], 'login_required')
        self.assertNotEqual(json_data['data'], 'foo')

    def test_permission_dectorator_user(self):
        data = {'verb': 'test_perm'}
        connection = TestConnection(user=TestUser(username='testuser', id=1))
        self.permission_handler(connection).handle(data)
        json_data = json.loads(connection.sent_data[-1])
        self.assertEqual(json_data['context']['state'], 'success')
        self.assertEqual(json_data['data'], 'foo')



from datetime import datetime
from decimal import Decimal
from ..serializers.base_serializer import BaseSerializer
from ..serializers.serializer_importer import get_serializer, _imported_modules_
from .models import Company, Department, Staff, Document, FooModel, BarModel
from .mock_connection import TestConnection
from .serializers import DocumentSerializer, FooSerializer
from .dragon_django_test_case import DragonDjangoTestCase


class TestObject(object):
    def __init__(self, name, numval):
        super(TestObject, self).__init__()
        self.name = name
        self.numval = numval
        self.created = datetime.now()


class TestObjectSerializer(BaseSerializer):
    publish_fields = ['name', 'numval', 'created', 'dec_value']


class TestObjectWithSerializer(object):
    def __init__(self, name, numval):
        super(TestObjectWithSerializer, self).__init__()
        self.name = name
        self.numval = numval
        self.created = datetime.now()
        self.dec_value = Decimal(12.44)


class SerializerTest(DragonDjangoTestCase):
    def setUp(self):
        self.connection = TestConnection()

    def test_custom_serializer(self):
        to = TestObjectWithSerializer('foo', 12)
        data = TestObjectSerializer().serialize(to)
        self.assertEqual('foo', data['name'])
        self.assertEqual(12, data['numval'])
        self.assertTrue(isinstance(data['created'], str))
        self.assertTrue(isinstance(data['dec_value'], str))

    def test_m2m_serialization(self):
        with Company() as company:
            company.name = 'company a'
            company.comp_num = 33
            company.pk = 321

        with Department() as department:
            department.name = 'dep a'
            department.company = company

        with Document() as document:
            document.name = 'test doc'

        with Staff() as staff:
            staff.name = 'John Doe'
            staff.department = department
            staff.pk = 123

        staff.documents.add(document)
        doc_ser = DocumentSerializer(instance=document).serialize()
        self.assertEqual(staff.pk, doc_ser['staff'][0]['id'])

    def test_get_serializer_from_string(self):
        ser = get_serializer('tests.DocumentSerializer', self.__class__)
        self.assertEqual(ser, DocumentSerializer)
        self.assertNotEqual(_imported_modules_, {})

    def test_serialize_include_via(self):
        '''
        Ensure serializing related models includes the 'via' field
        to help with the object mapper
        '''
        foo = FooModel.objects.create(test_field_a='test')
        bar = BarModel.objects.create(number=123, foo=foo)
        data = FooSerializer(instance=foo).serialize()
        self.assertEqual(data['bars'][0]['foo_id'], foo.pk)



from ..pubsub_providers.base_provider import BaseProvider

_channels = []
_subscribers = {}


class MockPubSubProvider(BaseProvider):
    def __init__(self):
        self._channels = _channels
        self._subscribers = _subscribers

    def get_channel(self, base_channel, **channel_filter):
        return self._construct_channel(base_channel, **channel_filter)

    def _get_channels_starting_with(self, base_channel):
        channels = []
        for c in self._channels:
            if str(c).startswith(base_channel):
                channels.append(c)
        return channels

    def get_channels(self, base_channel):
        return self._get_channels_starting_with(base_channel)

    def publish(self, channel, data):
        connections = self._subscribers.get(channel, [])
        for c in connections:
            c.publish(data)

    def subscribe(self, channels, connection):
        if not isinstance(channels, list):
            channels = [channels, ]
        for c in channels:
            if c not in self._channels:
                self._channels.append(c)
            if c not in self._subscribers.keys():
                self._subscribers[c] = []
            if connection not in self._subscribers[c]:
                self._subscribers[c].append(connection)

    def unsubscribe(self, channels, connection):
        for c in channels:
            if c in self._subscribers and self._subscribers[c] is not None:
                self._subscribers[c].remove(connection)
                if not self._subscribers[c]:
                    self._subscribers.pop(c)



from ..pubsub_providers.model_channel_builder import make_channels, filter_channels_by_model
from .serializers import CompanySerializer, DepartmentSerializer, LogoSerializer, StaffSerializer, DocumentSerializer
from ..pubsub_providers.channel_utils import channel_match_check
from .dragon_test_case import DragonTestCase
from .models import Company, Department


class TestChannelConstructor(DragonTestCase):
    def test_channels_from_company(self):
        property_filter = {'name__contains': 'foo'}
        channels = make_channels(CompanySerializer, [DepartmentSerializer, LogoSerializer], {'name__contains': 'foo'})
        expecpted = [
            'company|name__contains:foo',
            'department|company__name__contains:foo',
            'companylogo|company__name__contains:foo'
        ]
        self.assertListEqual(channels, expecpted)

    def test_channels_from_companylogo(self):
        property_filter = {'url': 'www.test.com'}
        channels = make_channels(LogoSerializer, property_filter=property_filter)
        expecpted = [
            'companylogo|url:www.test.com',
        ]
        self.assertListEqual(channels, expecpted)

    def test_channels_from_department(self):
        channels = make_channels(DepartmentSerializer, [StaffSerializer], {'name__contains': 'foo'})
        expecpted = [
            'department|name__contains:foo',
            'staff|department__name__contains:foo'
        ]
        self.assertListEqual(channels, expecpted)

    def test_channels_from_staff(self):
        property_filter = {'name__contains': 'foo'}
        channels = make_channels(StaffSerializer, [DocumentSerializer], property_filter)
        expected = [
            'staff|name__contains:foo',
            'document|staff__name__contains:foo'
        ]
        self.assertListEqual(channels, expected)

    def test_channels_from_document(self):
        property_filter = {'content__contains': 'foo'}
        channels = make_channels(DocumentSerializer, [StaffSerializer], property_filter)
        expecpted = [
            'document|content__contains:foo',
            'staff|document__content__contains:foo',
        ]
        self.assertListEqual(channels, expecpted)


class TestChannelUtils(DragonTestCase):
    def test_channel_match_check(self):
        channel = u'department|company__name__contains:foo'
        data = {'company_id': 1, 'id': 1, '_type': 'department', 'name': 'updated', 'staff': []}
        channel_match_check(channel, data)

    def test_filter_channels_by_related_in(self):
        comp = Company.objects.create(name='test co', comp_num=123)
        dep_a = Department.objects.create(company=comp, name='dep a')

        channels = [
            'company|departments__id__in:[3, {}, 99999]'.format(dep_a.pk)
        ]
        res = filter_channels_by_model(channels, comp)
        self.assertEqual(res, channels)

    def test_filter_channels_by_related_multiple_in(self):
        comp = Company.objects.create(name='test co', comp_num=123)
        dep_a = Department.objects.create(company=comp, name='dep a')

        channels = [
            'company|departments__id__in:[3, {}, 99999]|name:test co'.format(dep_a.pk)
        ]
        res = filter_channels_by_model(channels, comp)
        self.assertEqual(res, channels)



from django.db import models


class ParentModel(models.Model):
    name = models.CharField(max_length=100)
    age = models.PositiveIntegerField()


class ChildModel(models.Model):
    parent = models.ForeignKey(ParentModel, related_name='children')
    name = models.CharField(max_length=100)


class SubChildModel(models.Model):
    child = models.ForeignKey(ChildModel, related_name='subchildren')
    name = models.CharField(max_length=100)



from .serializers import FooSerializer, BarSerializer, BazSerializer, QuxSerializer, StaffSerializer, DocumentSerializer
from .models import FooModel, BarModel, BazModel, QuxModel, Company, Document, Staff
from ..tests.dragon_django_test_case import DragonDjangoTestCase


class DeserializerTest(DragonDjangoTestCase):
    def test_meta(self):
        serializer = FooSerializer()
        self.assertIn('test_field_a', serializer.opts.publish_fields)
        self.assertIn('test_field_a', serializer.opts.update_fields)
        self.assertIn('test_field_b', serializer.opts.update_fields)

    def test_deserialize_simple_model(self):
        """
        Deserialize a simple model with no relationships
        """
        data = {'test_field_a': 'foo', 'test_field_b': 'bar'}
        foo = FooSerializer(data=data).save()
        self.assertEqual('foo', foo.test_field_a)
        self.assertEqual('bar', foo.test_field_b)

    def test_deserialize_with_fk(self):
        """
        Deserialize a model with a foreign key
        """
        data = {
            'number': 12,
            'foo': {'test_field_a': 'foo', 'test_field_b': 'bar'}
        }
        bar = BarSerializer(data).save()
        self.assertEqual(bar.number, 12)
        self.assertEqual(bar.foo.test_field_a, 'foo')

    def test_deserialize_reverse_fks(self):
        """
        Deserialize a model with a list of reverse FKs
        """
        data = {
            'test_field_a': 'foo',
            'test_field_b': 'bar',
            'bars': [
                {'number': 52, },
                {'number': 42, },
            ]
        }
        serializer = FooSerializer(data)
        foo = serializer.save()
        self.assertTrue(foo.bars.exists())

    def test_deserialize_one_to_one(self):
        """
        Deserialize a model with a one to one relationship
        """
        data = {
            'name': 'this is baz',

            'bar': {'number': 25}
        }
        baz = BazSerializer(data).save()
        self.assertEqual(baz.name, 'this is baz')
        self.assertEqual(baz.bar.number, 25)

    def test_deserialize_m2m(self):
        data = {
            'test_field_a': 'foo',
            'test_field_b': 'bar',
            'value': 'test',
            'foos': [{'test_field_a': 'a', 'test_field_b': 'b'}]
        }
        qux = QuxSerializer(data).save()
        self.assertTrue(qux.foos.exists())


class SerializerTest(DragonDjangoTestCase):
    def test_serialize_simple_model(self):
        """
        Serialize a model with no FKs
        """
        foo = FooModel.objects.create(test_field_a='hello', test_field_b='world')
        serializer = FooSerializer(instance=foo)
        data = serializer.serialize()
        for pub_field in serializer.opts.publish_fields:
            self.assertIn(pub_field, data)

    def test_serialize_with_fk(self):
        """
        Serialize a model with a foreign key
        """
        foo = FooModel.objects.create(test_field_a='hello', test_field_b='world')
        bar = BarModel.objects.create(number=123, foo=foo)
        serializer = BarSerializer(instance=bar)
        data = serializer.serialize()
        self.assertEqual(data['foo']['test_field_a'], foo.test_field_a)

    def test_serialize_reverse_fks(self):
        """
        Serialize a model with a reverse foreign key.
        Make sure infinite recursion does not occur
        """
        foo = FooModel.objects.create(test_field_a='hello', test_field_b='world')
        foo.bars.add(BarModel.objects.create(number=123))
        foo.bars.add(BarModel.objects.create(number=333))
        serializer = FooSerializer(instance=foo)
        data = serializer.serialize()

    def test_serialize_one_to_one(self):
        """
        Serialize a model with a one to one relationship
        """
        bar = BarModel.objects.create(number=333)
        baz = BazModel.objects.create(name='bazzy', bar=bar)
        serializer = BazSerializer(instance=baz)
        data = serializer.serialize()
        self.assertEqual(data['bar']['number'], 333)
        self.assertEqual(data['name'], 'bazzy')

    def test_serialize_m2m(self):
        """
        Serialize a model with m2m fields
        """
        qux = QuxModel.objects.create(value='qux')
        qux.foos.add(FooModel.objects.create(test_field_a='a', test_field_b='b'))
        serializer = QuxSerializer(instance=qux)
        data = serializer.serialize()
        self.assertEqual(data['value'], 'qux')
        self.assertEqual(data['foos'][0]['test_field_a'], 'a')


class SerializerExtraDataTest(DragonDjangoTestCase):
    '''
    Test that serializers include mappings for ids and _type values
    '''
    def test_serialize_contains_type(self):
        '''
        Ensure that the serializer includes the '_type' property
        as required by the data mapper
        '''
        foo_ser = FooSerializer(instance=FooModel.objects.create(test_field_a='hello', test_field_b='world'))
        data = foo_ser.serialize()
        self.assertEqual(data['_type'], 'foomodel')

    def test_serialize_contains_list_of_related_ids_reverse_m2m(self):
        '''
        Ensure that the serializer include related objects ids
        on a reverse many 2 many relation
        '''
        staff = Staff.objects.create(name='staffy duck')
        document = Document.objects.create(content='test 123')
        staff.documents.add(document)
        ser = StaffSerializer(instance=staff)
        data = ser.serialize()
        self.assertListEqual([staff.pk], data['documents'][0]['staff_id'])

    def test_serialize_contains_list_of_related_ids_m2m(self):
        '''
        Ensure that the serializer include related objects ids
        on a many 2 many relation
        '''
        staff = Staff.objects.create(name='staffy duck')
        document = Document.objects.create(content='test 123')
        staff.documents.add(document)
        ser = DocumentSerializer(instance=document)
        data = ser.serialize()
        self.assertListEqual([document.pk], data['staff_id'])

    def test_serialize_contains_list_of_related_ids_related(self):
        '''
        Ensure that the serializer include related objects ids
        on a many 2 many relation
        '''
        foo = FooModel.objects.create(test_field_a='hello', test_field_b='world', pk=100)
        bar = BarModel.objects.create(number=123, foo=foo)
        ser = BarSerializer(instance=bar)
        data = ser.serialize()
        self.assertListEqual([foo.pk], data['foo_id'])



from ..route_handler import BaseModelPublisherRouter
from .models import Company
from .serializers import CompanySerializer
from .dragon_django_test_case import DragonDjangoTestCase
from swampdragon import route_handler


class CompanyRouter(BaseModelPublisherRouter):
    model = Company
    route_name = 'test_company_router'
    serializer_class = CompanySerializer
    paginate_by = 3

    def get_query_set(self, **kwargs):
        return self.model.objects.filter(**kwargs)

    def get_object(self, **kwargs):
        return self.model.objects.get(**kwargs)


class TestPagination(DragonDjangoTestCase):
    def setUp(self):
        route_handler.register(CompanyRouter)

        for i in range(10):
            Company.objects.create(name='test co', comp_num=i)

    def test_paginate_result(self):
        self.connection.call_verb(CompanyRouter.route_name, 'get_list', **{'_page': 1})
        page_1 = self.connection.get_last_message()['data']
        self.connection.call_verb(CompanyRouter.route_name, 'get_list', **{'_page': 2})
        page_2 = self.connection.get_last_message()['data']
        self.connection.call_verb(CompanyRouter.route_name, 'get_list', **{'_page': 3})
        page_3 = self.connection.get_last_message()['data']
        self.connection.call_verb(CompanyRouter.route_name, 'get_list', **{'_page': 4})
        page_4 = self.connection.get_last_message()['data']

        self.assertListEqual(
            [p['id'] for p in page_1],
            [c['pk'] for c in Company.objects.all()[0:3].values('pk')]
        )
        self.assertListEqual(
            [p['id'] for p in page_2],
            [c['pk'] for c in Company.objects.all()[3:6].values('pk')]
        )
        self.assertListEqual(
            [p['id'] for p in page_3],
            [c['pk'] for c in Company.objects.all()[6:9].values('pk')]
        )
        self.assertListEqual(
            [p['id'] for p in page_4],
            [c['pk'] for c in Company.objects.all()[9:].values('pk')]
        )

    def test_paginator(self):
        self.connection.call_verb(CompanyRouter.route_name, 'get_list', **{'_page': 1})
        page_1 = self.connection.get_last_message()['context']['client_context']
        self.assertTrue(page_1['page']['has_next'])
        self.assertFalse(page_1['page']['has_previous'])

        self.connection.call_verb(CompanyRouter.route_name, 'get_list', **{'_page': 2})
        page_2 = self.connection.get_last_message()['context']['client_context']
        self.assertTrue(page_2['page']['has_next'])
        self.assertTrue(page_2['page']['has_previous'])

        self.connection.call_verb(CompanyRouter.route_name, 'get_list', **{'_page': 4})
        page_4 = self.connection.get_last_message()['context']['client_context']
        self.assertFalse(page_4['page']['has_next'])
        self.assertTrue(page_4['page']['has_previous'])



from .dragon_django_test_case import DragonDjangoTestCase
from ..serializers.serializer_tools import get_serializer_relationship_field
from .serializers import FooSerializer, BarSerializer, StaffSerializer, DocumentSerializer


class SerializerTest(DragonDjangoTestCase):
    def test_get_serializer_relationship_m2m(self):
        rel = get_serializer_relationship_field(DocumentSerializer, StaffSerializer)
        self.assertEqual(rel, 'document')

    def test_get_serializer_relationship_reverse_m2m(self):
        rel = get_serializer_relationship_field(StaffSerializer, DocumentSerializer)
        self.assertEqual(rel, 'staff')

    def test_get_serializser_relationship_with_reverse_fk(self):
        rel = get_serializer_relationship_field(FooSerializer, BarSerializer)
        self.assertEqual(rel, 'foo')

    def test_get_serializser_relationship_with_fk(self):
        rel = get_serializer_relationship_field(BarSerializer, FooSerializer)
        self.assertEqual(rel, 'barmodel')



from django.db import models
from ..models import SelfPublishModel
from .mock_provider import MockPubSubProvider
from .serializers import CompanySerializer, DepartmentSerializer, StaffSerializer, DocumentSerializer, LogoSerializer, \
    FooSerializer, BarSerializer, CacheFooSerializer, CacheBarSerializer
from ..serializers.model_serializer import ModelSerializer


class TestSelfPublishModel(SelfPublishModel):
    publisher_class = MockPubSubProvider


class CompanyLogo(TestSelfPublishModel, models.Model):
    url = models.CharField(max_length=100)
    serializer_class = LogoSerializer


class Company(TestSelfPublishModel, models.Model):
    serializer_class = CompanySerializer
    name = models.CharField(max_length=100)
    comp_num = models.PositiveIntegerField()
    logo = models.OneToOneField(CompanyLogo, null=True)


class Department(TestSelfPublishModel, models.Model):
    serializer_class = DepartmentSerializer
    name = models.CharField(max_length=100)
    company = models.ForeignKey(Company, related_name='departments')


class Staff(TestSelfPublishModel, models.Model):
    serializer_class = StaffSerializer
    name = models.CharField(max_length=100)
    department = models.ForeignKey(Department, related_name='staff', null=True)


class Document(TestSelfPublishModel, models.Model):
    name = models.CharField(max_length=100)
    content = models.TextField(default='test doc')
    staff = models.ManyToManyField(Staff, related_name='documents')
    serializer_class = DocumentSerializer


class FooWithAbstractSerializer(ModelSerializer):
    class Meta:
        model = 'tests.FooWithAbstractBase'
        publish_fields = ['name']
        update_fields = ['name']


class BarWithAbstractSerializer(ModelSerializer):
    class Meta:
        model = 'tests.BarWithAbstractBase'
        publish_fields = ['name', 'is_something']
        update_fields = ['name', 'is_something']


class FooAbstract(TestSelfPublishModel, models.Model):
    class Meta:
        abstract = True


class FooWithAbstractBase(FooAbstract):
    serializer_class = FooWithAbstractSerializer
    name = models.CharField(max_length=100)

    def save(self, *args, **kwargs):
        self.name = 'override: {}'.format(self.name)
        super(FooWithAbstractBase, self).save(*args, **kwargs)


class BarWithAbstractBase(FooAbstract):
    name = models.CharField(max_length=100)
    is_something = models.BooleanField(default=True)

    def save(self, *args, **kwargs):
        self.name = 'override: {}'.format(self.name)
        super(BarWithAbstractBase, self).save(*args, **kwargs)


class TestUser(object):
    def __init__(self, username=None, id=None):
        self.username = username
        self.id = id

    def is_anonymous(self):
        return self.id is None


class FooModel(TestSelfPublishModel, models.Model):
    serializer_class = FooSerializer
    test_field_a = models.CharField(max_length=100)
    test_field_b = models.CharField(max_length=100)


class BarModel(TestSelfPublishModel, models.Model):
    serializer_class = BarSerializer
    number = models.IntegerField()
    foo = models.ForeignKey(FooModel, related_name='bars', null=True, blank=True)


class BazModel(models.Model):
    name = models.CharField(max_length=100)
    bar = models.OneToOneField(BarModel, null=True, blank=True)


class QuxModel(models.Model):
    value = models.CharField(max_length=100)
    foos = models.ManyToManyField(FooModel)


class CacheFooModel(TestSelfPublishModel, models.Model):
    serializer_class = CacheFooSerializer
    test_field_a = models.CharField(max_length=100)
    test_field_b = models.CharField(max_length=100)


class CacheBarModel(TestSelfPublishModel, models.Model):
    serializer_class = CacheBarSerializer
    number = models.IntegerField()
    foo = models.ForeignKey(CacheFooModel, related_name='bars', null=True)



from .. import route_handler
from ..route_handler import BaseModelPublisherRouter
from .models import Company, CompanySerializer, FooWithAbstractBase, FooWithAbstractSerializer, BarWithAbstractBase, \
    BarWithAbstractSerializer, Department, Staff, Document
from .mock_connection import TestConnection
from .serializers import DocumentSerializer, DepartmentSerializer, StaffSerializer
from .dragon_django_test_case import DragonDjangoTestCase


class DocumentRouter(BaseModelPublisherRouter):
    model = Document
    route_name = 'test_doc_router'
    serializer_class = DocumentSerializer

    def get_query_set(self, **kwargs):
        return self.model.objects.filter(**kwargs)

    def get_object(self, **kwargs):
        return self.model.objects.get(**kwargs)


class CompanyRouter(BaseModelPublisherRouter):
    model = Company
    route_name = 'test_company_router'
    serializer_class = CompanySerializer
    include_related = [DepartmentSerializer]

    def get_query_set(self, **kwargs):
        return self.model.objects.filter(**kwargs)

    def get_object(self, **kwargs):
        return self.model.objects.get(**kwargs)


class FooWithAbstractRouter(BaseModelPublisherRouter):
    model = FooWithAbstractBase
    route_name = 'test_foo_abs_router'
    serializer_class = FooWithAbstractSerializer

    def get_query_set(self, **kwargs):
        return self.model.objects.filter(**kwargs)

    def get_object(self, **kwargs):
        return self.model.objects.get(**kwargs)


class BarWithAbstractRouter(BaseModelPublisherRouter):
    model = BarWithAbstractBase
    route_name = 'test_foo_abs_router'
    serializer_class = BarWithAbstractSerializer

    def get_query_set(self, **kwargs):
        return self.model.objects.filter(**kwargs)

    def get_object(self, **kwargs):
        return self.model.objects.get(**kwargs)


class TestSelfPubModel(DragonDjangoTestCase):
    def setUp(self):
        route_handler.register(CompanyRouter)
        route_handler.register(FooWithAbstractRouter)
        route_handler.register(DocumentRouter)
        self.connection = TestConnection()
        self.company_handler = route_handler.get_route_handler(CompanyRouter.route_name)
        self.foo_abs_handler = route_handler.get_route_handler(FooWithAbstractRouter.route_name)
        self.bar_abs_handler = route_handler.get_route_handler(BarWithAbstractRouter.route_name)
        self.document_handler = route_handler.get_route_handler(DocumentRouter.route_name)

    def test_self_publish(self):
        kwargs = {'channel': 'client_chan', }
        self.company_handler(self.connection).subscribe(**kwargs)
        foo = Company.objects.create(name='foo', comp_num=33)
        json_data = self.connection.get_last_published_data()
        serialized_foo = foo.serialize()
        self.assertDictEqual(serialized_foo, json_data)

    def test_self_publish_with_abstract(self):
        kwargs = {'channel': 'client_chan', }
        self.foo_abs_handler(self.connection).subscribe(**kwargs)
        foo = FooWithAbstractBase.objects.create(name='foo')
        json_data = self.connection.get_last_published_data()
        serialized_foo = foo.serialize()
        self.assertDictEqual(serialized_foo, json_data)

    def test_self_publish_with_abstract_without_serializer(self):
        '''
        Assure it raises an exception, as the BarWithAbstractBase does not have
        a serializer class
        '''
        kwargs = {'channel': 'client_chan', }
        self.bar_abs_handler(self.connection).subscribe(**kwargs)
        with self.assertRaises(TypeError):
            BarWithAbstractBase.objects.create(name='foo')

    def test_only_publish_changes(self):
        kwargs = {'channel': 'client_chan', }
        self.company_handler(self.connection).subscribe(**kwargs)
        foo = Company.objects.create(name='foo', comp_num=33)

        self.connection.sent_data = []
        foo.comp_num = 33
        foo.save()
        self.assertEqual(len(self.connection.sent_data), 0)

        foo.comp_num = 34
        foo.save()
        data = self.connection.get_last_published_data()
        self.assertEqual(data['comp_num'], foo.comp_num)

    def test_publish_based_on_children(self):
        with Company(name='foo company', comp_num=33) as company:
            pass
        with Department(name='department a', company=company) as department:
            pass
        self.connection.subscribe('test_company_router', 'cli_chan', {'name__contains': 'foo'})

        department.name = 'updated'
        department.save()
        self.assertGreater(len(self.connection.published_data), 0)

    def test_no_publish_based_on_childrens_children(self):
        kwargs = {'channel': 'client_chan', 'id': 1}
        company = Company.objects.create(name='company a', comp_num=33)
        department = Department.objects.create(name='department a', company=company)
        Staff.objects.create(name='John', department=department)
        dont_publish_staff = Staff.objects.create(
            name='no pub staff',
            department=Department.objects.create(name='foo', company=Company.objects.create(name='foo2', comp_num=33)),
        )
        self.company_handler(self.connection).subscribe(**kwargs)
        self.connection.sent_data = []
        dont_publish_staff.name = 'updated'
        dont_publish_staff.save()
        self.assertEqual(len(self.connection.sent_data), 0)

    def test_get_changes(self):
        kwargs = {'channel': 'client_chan', }
        self.company_handler(self.connection).subscribe(**kwargs)
        foo = Company.objects.create(name='foo', comp_num=33)
        expected = foo._get_changes()
        foo.comp_num = 33
        actual = foo._get_changes()
        self.assertDictEqual(expected, actual)

    def test_m2m(self):
        kwargs = {'channel': 'client_chan'}
        self.document_handler(self.connection).subscribe(**kwargs)
        with Staff(name='John') as staff_a:
            pass
        with Staff(name='Tina') as staff_b:
            pass

        self.connection.sent_data = []

        with Document(name='test doc') as document:
            document.save()
            document.staff.add(staff_a)
            document.staff.add(staff_b)
        document.name = 'test doc updated'
        document.save()
        self.assertEqual(len(self.connection.published_data), 3)

    def test_remove_on_update(self):
        with Company(name='foo', comp_num=55) as company:
            pass
        kwargs = {'channel': 'client_chan', 'name': 'foo'}
        self.company_handler(self.connection).subscribe(**kwargs)
        company.name = 'updated'
        company.save()
        last_update = self.connection.get_last_published()
        self.assertEqual(last_update['action'], 'deleted')



from .dragon_test_case import DragonTestCase
from ..sessions.mock_session_store import MockSessionStore


class SessionTest(DragonTestCase):
    def setUp(self):
        self.session_store = MockSessionStore()

    def test_save_sessions(self):
        key = 'foo'
        val = 'bar'

        self.session_store.save(key, val)
        actual = self.session_store.get(key)
        self.assertEqual(val, actual)

    def test_save_sessions_get_key(self):
        val = 'bar'
        key = self.session_store.save_get_key(val)
        actual = self.session_store.get(key)
        self.assertEqual(val, actual)



from sockjs.tornado import SockJSConnection
from ..pubsub_providers.redis_pubsub_provider import RedisPubSubProvider
from .. import route_handler
import json


class ConnectionMixin(object):
    def to_json(self, data):
        if isinstance(data, dict):
            return data
        try:
            data = json.loads(data.replace("'", '"'))
            return data
        except:
            return json.dumps({'message': data})

    def to_string(self, data):
        if isinstance(data, dict):
            return json.dumps(data).replace("'", '"')
        return data


class SubscriberConnection(ConnectionMixin, SockJSConnection):
    def __init__(self, session):
        super(SubscriberConnection, self).__init__(session)

    def on_open(self, request):
        self.pub_sub = RedisPubSubProvider()

    def on_close(self):
        self.pub_sub.close()

    def on_message(self, data):
        try:
            data = self.to_json(data)
            handler = route_handler.get_route_handler(data['route'])
            handler(self).handle(data)
        except Exception as e:
            self.abort_connection()
            raise e

    def abort_connection(self):
        self.close()

    def send(self, message, binary=False):
        super(SubscriberConnection, self).send(self.to_string(message), binary)

    def broadcast(self, clients, message):
        data = self.to_string(message)
        super(SubscriberConnection, self).broadcast(clients, data)


class DjangoSubscriberConnection(SubscriberConnection):
    def __init__(self, session):
        super(DjangoSubscriberConnection, self).__init__(session)



import json
from .. import route_handler
from ..route_handler import BaseModelPublisherRouter, UnexpectedVerbException
from .mock_connection import TestConnection
from .dragon_django_test_case import DragonDjangoTestCase
from .models import FooModel, BarModel
from .serializers import FooSerializer, BarSerializer


class FooRouter(BaseModelPublisherRouter):
    valid_verbs = BaseModelPublisherRouter.valid_verbs + ['get_misc_data']
    model = FooModel
    route_name = 'test_parent_router'
    serializer_class = FooSerializer
    include_related = [BarSerializer]

    def get_query_set(self, **kwargs):
        return self.model.objects.filter(**kwargs)

    def get_object(self, **kwargs):
        return self.model.objects.get(id=kwargs['id'])

    def get_misc_data(self, **kwargs):
        self.send({'foo': 'bar'})


class TestModelRouter(DragonDjangoTestCase):
    def setUp(self):
        route_handler.register(FooRouter)
        self.connection = TestConnection()
        self.foo_router = route_handler.get_route_handler(FooRouter.route_name)

    def test_subscribe(self):
        '''
        Subscribe to a channel
        '''
        data = {'verb': 'subscribe', 'args': {'channel': 'testchan'}}
        self.foo_router(self.connection).handle(data)
        data = self.connection.get_last_message()['data']

        self.assertListEqual(
            data,
            self.foo_router.serializer_class.get_object_map(self.foo_router.include_related)
        )

        remote_channels = json.loads(self.connection.sent_data[0])['channel_data']['remote_channels']
        for channel in remote_channels:
            self.assertTrue(self.connection in self.connection.pub_sub._subscribers[channel])

    def test_unsubscribe(self):
        '''
        Unsubscribe to a channel
        '''
        data = {'verb': 'subscribe', 'args': {'channel': 'testchan'}}
        self.foo_router(self.connection).handle(data)
        data = {'verb': 'unsubscribe', 'args': {'channel': 'testchan'}}
        self.foo_router(self.connection).handle(data)

        received_data = self.connection.get_last_message()

        self.assertEqual(received_data['data'], 'unsubscribed')
        remote_channels = received_data['channel_data']['remote_channels']
        for channel in remote_channels:
            self.assertTrue(self.connection not in self.connection.pub_sub._subscribers.get(channel, []))

    def test_create_model(self):
        model_data = {'test_field_a': 'test data', 'test_field_b': 'foo test'}
        data = {'verb': 'create', 'args': model_data}
        self.foo_router(self.connection).handle(data)
        self.assertTrue(self.foo_router.model.objects.filter(**model_data).exists())

    def test_update_model(self):
        model_data = {'test_field_a': 'value', 'test_field_b': 'second value'}
        foo = FooModel.objects.create(**model_data)

        update_data = {
            'id': foo.pk,
            'test_field_a': 'updated data'
        }

        data = {'verb': 'update', 'args': update_data}
        self.foo_router(self.connection).handle(data)
        updated_model = self.foo_router.model.objects.get(**{'pk': foo.pk})
        self.assertEqual(foo.pk, updated_model.pk)
        self.assertEqual(updated_model.test_field_a, 'updated data')

    def test_delete_model(self):
        foo_model = FooModel.objects.create(**{'test_field_a': 'test parent', 'test_field_b': '55'})
        data = {'verb': 'delete', 'args': {'id': foo_model.id}}
        self.foo_router(self.connection).handle(data)
        self.assertFalse(self.foo_router.model.objects.filter(**{'id': foo_model.id}).exists())

    def test_get_models(self):
        model_data = {'test_field_a': 'test', 'test_field_b': 'testing'}
        foo = FooModel.objects.create(**model_data)
        data = {'verb': 'get_list', 'args': {'id': foo.id}}
        self.foo_router(self.connection).handle(data)
        actual_data = self.connection.get_last_message()['data'][0]
        expected_data = FooSerializer(instance=foo).serialize()
        for ed in actual_data.keys():
            self.assertEqual(expected_data[ed], actual_data[ed])

    def test_get_model(self):
        foo = FooModel.objects.create(**{'test_field_a': 'test', 'test_field_b': '123 test'})
        data = {'verb': 'get_single', 'args': {'id': foo.id}}
        self.foo_router(self.connection).handle(data)
        expected_data = FooSerializer(instance=foo).serialize()
        actual_data = self.connection.get_last_message()['data']
        for ed in expected_data.keys():
            self.assertEqual(expected_data[ed], actual_data[ed])

    def test_unexpected_verb_exception(self):
        with self.assertRaises(UnexpectedVerbException):
            data = {'verb': 'madeupverb'}
            self.foo_router(self.connection).handle(data)

    def test_get_parents_with_children(self):
        foo = FooModel.objects.create(test_field_a='hello', test_field_b='world',)
        BarModel.objects.create(number=12, foo=foo)

        data = {'verb': 'get_list'}
        self.foo_router(self.connection).handle(data)
        received_data = self.connection.get_last_message()['data']
        self.assertEqual(len(received_data[0]['bars']), 1)

    def test_get_parent_with_children(self):
        foo = FooModel.objects.create(test_field_a='hello', test_field_b='world',)
        BarModel.objects.create(number=12, foo=foo)

        data = {'verb': 'get_single', 'args': {'id': foo.pk}}
        self.foo_router(self.connection).handle(data)
        received_data = self.connection.get_last_message()['data']
        self.assertEqual(len(received_data['bars']), 1)

    def test_get_models_with_filter(self):
        FooModel.objects.create(test_field_a='hello', test_field_b='world',)
        foo = FooModel.objects.create(test_field_a='findme', test_field_b='world',)
        data = {'verb': 'get_list', 'args': {'test_field_a__contains': 'find'}}
        self.foo_router(self.connection).handle(data)
        received_data = self.connection.get_last_message()
        self.assertTrue(len(received_data), 1)
        self.assertEqual(received_data['data'][0]['test_field_a'], foo.test_field_a)

    def test_subscribe_to_related(self):
        '''
        Test channels from related serializers are included
        '''
        foo = FooModel.objects.create(test_field_a='hello', test_field_b='world',)
        kwargs = {'channel': 'client_chan', 'id': foo.pk, }
        self.foo_router(self.connection).subscribe(**kwargs)
        expected_channels = [
            '{}id:1'.format(FooSerializer.get_base_channel()),
            '{}foo__id:1'.format(BarSerializer.get_base_channel()),
        ]
        data = self.connection.get_last_message()
        remote_channels = data['channel_data']['remote_channels']
        self.assertListEqual(remote_channels, expected_channels)

    def test_custom_get_method(self):
        FooModel.objects.create(test_field_a='hello', test_field_b='world',)
        FooModel.objects.create(test_field_a='foo', test_field_b='bar',)
        FooModel.objects.create(test_field_a='baz', test_field_b='qux',)
        data = {'verb': 'get_misc_data'}
        self.foo_router(self.connection).handle(data)
        data = self.connection.get_last_message()['data']
        self.assertDictEqual(data, {'foo': 'bar'})

    def test_remove_on_update(self):
        '''
        Causing a change in a model where it no longer matches a subscription
        should remove the model from the subscriber
        '''
        foo = FooModel.objects.create(test_field_a='hello', test_field_b='world',)
        kwargs = {'channel': 'client_chan', 'test_field_a': 'hello', }
        self.foo_router(self.connection).subscribe(**kwargs)
        self.foo_router(self.connection).update(**{'id': foo.pk, 'test_field_a': 'updated'})
        last_update = self.connection.get_last_published()
        self.assertEqual(last_update['action'], 'deleted')

    def test_specify_related_models_on_subscribe(self):
        self.foo_router(self.connection).subscribe(**{'channel': 'bar', 'id': 1})
        obj_map = json.loads(self.connection.sent_data[-1])['data']
        self.assertEqual(len(obj_map), 2)
        self.assertEqual(len(self.connection.pub_sub._channels), 2)



from .dragon_django_test_case import DragonDjangoTestCase
from swampdragon import route_handler
from .models import FooModel, BarModel
from .test_route_handler import FooRouter


class SubscriptionTest(DragonDjangoTestCase):
    def setUp(self):
        route_handler.register(FooRouter)

    def test_subscribe_to_related_via_router(self):
        foo = FooModel.objects.create(test_field_a='foo a', test_field_b='foo b')
        bar = BarModel.objects.create(foo=foo, number=1)
        self.connection.subscribe(FooRouter.route_name, 'cli-chan', {'bars__number_gte': 10})
        bar.number = 110
        bar.save()
        # import ipdb;ipdb.set_trace()



from django.db.models.signals import pre_delete, m2m_changed
from django.dispatch.dispatcher import receiver
from .pubsub_providers.redis_pubsub_provider import RedisPubSubProvider
from .pubsub_providers.base_provider import PUBACTIONS
from .model_tools import get_property
from .pubsub_providers.model_publisher import publish_model


class SelfPublishModel(object):
    _ignore_changes_for = None
    _should_publish = True
    serializer_class = None
    publisher_class = RedisPubSubProvider

    def __enter__(self):
        self._should_publish = False
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.save()
        self._should_publish = True

    def __init__(self, *args, **kwargs):
        self._pre_save_state = dict()
        super(SelfPublishModel, self).__init__(*args, **kwargs)
        self._serializer = self.serializer_class(instance=self)
        self._set_ignored_fields()
        relevant_fields = self._get_relevant_fields()

        for field in relevant_fields:
            val = get_property(self, field)
            if val is None:
                self._pre_save_state[field] = val
                continue
            if hasattr(val, 'all'):
                val = val.all()
            self._pre_save_state[field] = val

    def _set_ignored_fields(self):
        if self.__class__._ignore_changes_for is None:
            self.__class__._ignore_changes_for = set()
            for f in self._serializer._get_related_fields():
                self._ignore_changes_for.add(f)
            for f in self._serializer._get_m2m_fields():
                self._ignore_changes_for.add(f)

    def _get_relevant_fields(self):
        update_fields = list(self._serializer.opts.update_fields)
        publish_fields = list(self._serializer.opts.publish_fields)
        relevant_fields = set(update_fields + publish_fields)

        if self._serializer.opts.id_field in relevant_fields:
            relevant_fields.remove(self._serializer.opts.id_field)
        return relevant_fields

    def _get_changes(self):
        changes = dict()
        for k, v in self._pre_save_state.items():
            val = get_property(self, k)
            if hasattr(val, 'all'):
                val = val.all()
                if v is None:
                    v = []
                diff = list(set(val).symmetric_difference(set(v)))
                if len(diff) > 0:
                    changes[k] = diff
            elif val != v:
                changes[k] = v
        return changes

    def serialize(self):
        return self._serializer.serialize()

    def _publish(self, action, changes=None):
        if not self.serializer_class:
            return
        publish_model(self, self._serializer, self.publisher_class(), action, changes)

    def save(self, *args, **kwargs):
        if not self.pk:
            self.action = PUBACTIONS.created
            self.changes = None
        else:
            self.action = PUBACTIONS.updated
            self.changes = self._get_changes()
        super(SelfPublishModel, self).save(*args, **kwargs)

        if self._should_publish:
            self._publish(self.action, self.changes)


@receiver(m2m_changed)
def _self_publish_model_m2m_change(sender, instance, action, **kwargs):
    if not isinstance(instance, SelfPublishModel):
        return
    if not hasattr(instance, 'action'):
        return

    if action is 'post_clear':
        instance.changes = instance._get_changes()
        instance._publish(instance.action, instance.changes)
    if action is 'post_add':
        instance.changes = instance._get_changes()
        instance._publish(instance.action, instance.changes)


@receiver(pre_delete)
def _self_publish_model_delete(sender, instance, **kwargs):
    if isinstance(instance, SelfPublishModel):
        instance._publish(PUBACTIONS.deleted)



import json
import redis
import tornadoredis.pubsub
import tornadoredis
from .base_provider import BaseProvider


class RedisPubSubProvider(BaseProvider):
    def __init__(self):
        self._client = redis.StrictRedis()
        self._async_client = tornadoredis.Client()
        self._subscriber = None

    def _get_subscriber(self):
        if not self._subscriber:
            self._subscriber = tornadoredis.pubsub.SockJSSubscriber(tornadoredis.Client())
        return self._subscriber

    def get_channels(self, base_channel):
        return self._get_channels_from_redis(base_channel)

    def close(self):
        if self._subscriber:
            self._subscriber.close()

    def get_channel(self, base_channel, **channel_filter):
        return self._construct_channel(base_channel, **channel_filter)

    def subscribe(self, channels, broadcaster):
        self._get_subscriber().subscribe(channels, broadcaster)

    def unsubscribe(self, channels, broadcaster):
        self._get_subscriber().unsubscribe(channels, broadcaster)

    def _get_channels_from_redis(self, base_channel):
        channels = self._client.execute_command('PUBSUB', 'channels', '{}*'.format(base_channel))
        return [c.decode() for c in channels]

    def publish(self, channel, data):
        self._client.publish(channel, json.dumps(data))



from .dragon_django_test_case import DragonDjangoTestCase
from swampdragon import route_handler
from swampdragon.serializers.validation import ModelValidationError
from swampdragon.tests.serializers import FooSerializer
from swampdragon.tests.test_route_handler import FooRouter


class ValidationTest(DragonDjangoTestCase):
    def setUp(self):
        route_handler.register(FooRouter)

    def test_trigger_validation_error(self):
        data = {'test_field_a': 'missing field b'}
        with self.assertRaises(ModelValidationError):
            FooSerializer(data=data).save()

    def test_router_validation_error(self):
        data = {'test_field_a': 'missing field b'}
        self.connection.call_verb('test_parent_router', 'create', **data)
        message = self.connection.get_last_message()
        self.assertEqual('error', message['context']['state'])



from ..serializers.cached_serializer import CacheModelSerializer
from ..serializers.model_serializer import ModelSerializer

################################
# Old school serializers
################################


class DepartmentSerializer(ModelSerializer):
    staff = 'tests.StaffSerializer'

    class Meta:
        model = 'tests.Department'
        publish_fields = ['name', 'company_id', 'staff']
        update_fields = ['name']

    def serialize_staff(self, obj=None, serializer=None, ignore_fields=[]):
        return [StaffSerializer().serialize(s, ignore_fields=['departments']) for s in obj.staff.all()]


class StaffSerializer(ModelSerializer):
    documents = 'tests.DocumentSerializer'

    class Meta:
        model = 'tests.Staff'
        publish_fields = ('name', 'documents')
        update_fields = ('name', )
        documents_serializer = 'tests.DocumentSerializer'


class LogoSerializer(ModelSerializer):
    class Meta:
        model = 'tests.CompanyLogo'
        publish_fields = ['url']


class CompanySerializer(ModelSerializer):
    departments = DepartmentSerializer
    logo = LogoSerializer

    class Meta:
        model = 'tests.Company'
        publish_fields = ['name', 'comp_num', 'departments', 'logo', 'custom_field']
        update_fields = ['name', 'comp_num']

    def serialize_custom_field(self, obj=None, serializer=None, ignore_fields=[]):
        return 'custom field'


class DocumentSerializer(ModelSerializer):
    staff = StaffSerializer

    class Meta:
        model = 'tests.Document'
        publish_fields = ['name', 'content', 'staff']
        staff_serializer = StaffSerializer

    # def serialize_staff(self, obj=None, serializer=None, ignore_fields=[]):
    #     return [serializer.serialize(o, ignore_fields=['documents']) for o in obj.staff.all()]


################################
# New serializeres
################################
class FooSerializer(ModelSerializer):
    bars = 'tests.BarSerializer'

    class Meta:
        publish_fields = ('test_field_a', 'bars')
        update_fields = ('test_field_a', 'test_field_b', 'bars')
        model = 'tests.FooModel'


class BarSerializer(ModelSerializer):
    foo = FooSerializer

    class Meta:
        model = 'tests.BarModel'
        update_fields = ('number', 'foo')
        publish_fields = ('number', 'foo')


class BazSerializer(ModelSerializer):
    bar = BarSerializer

    class Meta:
        model = 'tests.BazModel'
        update_fields = ('name', 'bar')
        publish_fields = ('name', 'bar')


class QuxSerializer(ModelSerializer):
    foos = FooSerializer

    class Meta:
        model = 'tests.QuxModel'
        publish_fields = ('value', 'foos')
        update_fields = ('value', 'foos')


class CacheFooSerializer(CacheModelSerializer):
    bars = 'tests.CacheBarSerializer'

    class Meta:
        model = 'tests.CacheFooModel'
        publish_fields = ('test_field_a', 'bars')
        update_fields = ('test_field_a', 'test_field_b', 'bars')


class CacheBarSerializer(CacheModelSerializer):
    foo = CacheFooSerializer

    class Meta:
        model = 'tests.CacheBarModel'
        update_fields = ('number', 'foo')
        publish_fields = ('number', 'foo')



from collections import namedtuple
from django.db.models.fields.related import ForeignKey
from django.db.models.related import RelatedObject
from django.db.models.fields.related import ManyToManyField


class FieldType(namedtuple('FieldType', 'field, model, fk, m2m')):
    '''
    Determine if a field is an m2m, reverse m2m, fk or reverse fk
    '''
    @property
    def is_m2m(self):
        return self.fk is False and self.m2m is True and isinstance(self.field, RelatedObject)

    @property
    def is_reverse_m2m(self):
        return self.fk is True and self.m2m is True and isinstance(self.field, ManyToManyField)

    @property
    def is_fk(self):
        return self.fk is True and self.m2m is False and isinstance(self.field, ForeignKey)

    @property
    def is_reverse_fk(self):
        return self.fk is False and self.m2m is False and isinstance(self.field, RelatedObject)


def get_serializer_relationship_field(serializer, related_serializer):
    if isinstance(serializer, type):
        model = serializer().opts.model
    else:
        model = serializer.opts.model
    if isinstance(related_serializer, type):
        related_model = related_serializer().opts.model
    else:
        related_model = related_serializer.opts.model

    for field_name in related_model._meta.get_all_field_names():
        field_type = FieldType(*related_model._meta.get_field_by_name(field_name))
        field = field_type.field

        # Foreign key
        if field_type.is_fk and field.rel.to is model:
            return field.verbose_name

        # Reverse foreign key
        if field_type.is_reverse_fk and field.model is model:
            return field.var_name

        # M2m fields
        if field_type.is_m2m and field.model is model:
            return field.var_name

        # Reverse m2m field
        if field_type.is_reverse_m2m and field.rel.to is model:
            return field.attname


def get_id_mappings(serializer):
    if not serializer.instance:
        return {}

    data = {}
    for field_name in serializer.opts.publish_fields:
        if not hasattr(serializer, field_name):
            continue
        serializable_field = getattr(serializer, field_name)
        if not hasattr(serializable_field, 'serialize'):
            continue

        field_type = FieldType(*serializer.opts.model._meta.get_field_by_name(field_name))
        field = field_type.field

        if field_type.is_fk:
            val = getattr(serializer.instance, field_name)
            if val:
                data['{}_id'.format(field.verbose_name)] = [val.pk]

        if field_type.is_reverse_fk:
            # Check if this is a one 2 one field first
            try:
                val = getattr(serializer.instance, field_name)
            except field.model.DoesNotExist:
                continue

            if hasattr(val, 'all'):
                qs = val.all()
                data['{}_id'.format(field.var_name)] = [v[0] for v in qs.values_list('pk')]

        if field_type.is_m2m:
            qs = getattr(serializer.instance, field_name).all()
            data['{}_id'.format(field.var_name)] = [v[0] for v in qs.values_list('pk')]

        if field_type.is_reverse_m2m:
            qs = getattr(serializer.instance, field_name).all()
            data['{}_id'.format(field.attname)] = [v[0] for v in qs.values_list('pk')]

    return data



# from django.utils.unittest.case import skip
# from .dragon_django_test_case import DragonDjangoTestCase
# from .models import CacheFooModel, CacheBarModel, FooModel, BarModel
#
#
# class SerializerExtraDataTest(DragonDjangoTestCase):
#     '''
#     Test caching of serializer data.
#     Serializing foo should result in no queries
#     '''
#     def test_lots_of_data(self):
#         model_count = 10
#
#         foo = FooModel.objects.create(test_field_a='hello', test_field_b='world')
#         for i in range(model_count):
#             bar = BarModel.objects.create(number=i, foo=foo)
#
#         for bar in BarModel.objects.all():
#             # with self.assertNumQueries(11):
#             data = foo.serialize()
#
#         foo = CacheFooModel.objects.create(test_field_a='hello', test_field_b='world')
#         for i in range(model_count):
#             bar = CacheBarModel.objects.create(number=i, foo=foo)
#
#         for bar in CacheBarModel.objects.all():
#             with self.assertNumQueries(0):
#                 data = foo.serialize()
#                 self.assertGreater(len(data['bars']), 0)



from ..pubsub_providers.base_provider import PUBACTIONS
from ..pubsub_providers.model_channel_builder import filter_channels_by_model, filter_channels_by_dict


def publish_model(model_instance, serializer, publisher, action, changes=None):
    if action is PUBACTIONS.updated and not changes:
        return

    if hasattr(serializer, '_cache'):
        serializer.remove_from_cache()

    base_channel = serializer.get_base_channel()
    all_model_channels = publisher.get_channels(base_channel)
    channels = filter_channels_by_model(all_model_channels, model_instance)

    serialized_data = serializer.serialize()
    if channels:
        # if changes:
        #     publish_data = dict({'data': {'id': model_instance.pk}})
        #     publish_data.update(changes)
        # else:
        publish_data = dict({'data': serialized_data})
        publish_data['action'] = action
        for c in channels:
            publish_data['channel'] = c
            publisher.publish(c, publish_data)

    if changes:
        previous_relevant_channels = filter_channels_by_dict(all_model_channels, changes)
        remove_from_channels = list(set(previous_relevant_channels) - set(channels))
        if not remove_from_channels:
            return
        publish_data = dict({'data': {'id': model_instance.pk}})
        publish_data['action'] = PUBACTIONS.deleted
        for c in remove_from_channels:
            publish_data['channel'] = c
            publisher.publish(c, publish_data)





def deserialize_related(serializer, model_instance, key, val):
    ser = getattr(serializer, '{}_serializer'.format(key))()
    if isinstance(val, list):
        for v in val:
            child = ser.deserialize(**v)
            getattr(model_instance, key).add(child)
